"""Use chain_list.csv to extract chains from PDB files.

This script perform the following:
1. Select only the first model from NMR structures
2. Save only chains of interest with headers from the original PDB
See:
- https://www.rcsb.org/docs/general-help/
identifiers-in-pdb#macromolecular-instance-id
"""
import pathlib
import csv
import subprocess

from Bio.PDB import PDBParser

parent = pathlib.Path(__file__).parent.resolve()
membrane_dir = pathlib.Path("/mnt/MEMBRANE/")
pdb_dir = membrane_dir / "PDB"
chain_dir = membrane_dir / "CHAINS"

chain_list_file = membrane_dir / "chains_list.csv"
with open(chain_list_file, "r", encoding="utf-8") as infile:
    reader = csv.reader(infile)
    chain_list = list(reader)

parser = PDBParser(QUITE=True)

for row in chain_list:
    pdb_path = str(pdb_dir / f"{row[0][:4]}.pdb")
    chain_id = row[2]  # auth_asym_id

    # a priori check if chain exists
    structure = parser.get_structure(pdb_path[-8:-4], pdb_path)
    models = list(structure.get_models())
    nmr = False
    if len(models) > 1:
        nmr = True
        print(f"{row[0][:4]}.pdb is NMR")
    chains = models[0]  # FIRST MODEL!!
    chain_found = False
    for ch in chains:
        if ch.id == chain_id:
            chain_found = True
            break
    if not chain_found:
        raise RuntimeError(f"Chain {chain_id} not found in {pdb_path}")

    # Extract chain (of the first model if NMR structure)
    # If NMR: extract first model and save it to temp file
    if nmr:
        cmd_model = ["pdb_selmodel", "-1", pdb_path]
        run_model = subprocess.run(
            cmd_model, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        if run_model.returncode != 0:
            raise RuntimeError(f"Model selection failed: {run_model.stdout}")
        out_model_file = chain_dir / f"{row[0][:4]}_1.pdb"
        with open(out_model_file, "w", encoding="utf-8") as outfile:
            outfile.write(run_model.stdout.decode("utf-8"))
    # Extract chain
    cmd_chain = ["pdb_selchain", f"-{chain_id}"]
    if nmr:
        cmd_chain += [str(chain_dir / f"{row[0][:4]}_1.pdb")]
    else:
        cmd_chain += [pdb_path]

    run_chain = subprocess.run(
        cmd_chain, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    if run_chain.returncode != 0:
        raise RuntimeError(f"Chain selection failed: {run_chain.stdout}")
    out_chain_file = chain_dir / f"{row[0][:4]}_{chain_id}.pdb"
    with open(out_chain_file, "w", encoding="utf-8") as outfile:
        outfile.write(run_chain.stdout.decode("utf-8"))
    if nmr:
        cmd_rm = ["rm", str(chain_dir / f"{row[0][:4]}_1.pdb")]
        run_rm = subprocess.run(
            cmd_rm, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        if run_rm.returncode != 0:
            raise RuntimeError(f"Removing temp file failed: {run_rm.stdout}")
