"""Model missing residues.

The scripts does the following:
1. Read REMARK 465 records to check if there are missing residues
and, in case, it stores them
2. take the full sequence from the SEQRES and align it with
the sequence obtained from the ATOM entries
3. Using Modeller to model those chains parts that are missing
(see the code for more details)
4. save chain with original residue numeration, both modeled and not

Note: the tails are modelled but will be removed
The rest of the chain remains unchanged because is kept fixed

Note: residues which are not involved into the missing region are
kept fixed. Except for the 3 residues before and after the missing
region. This is done to allow some mobility of the nearest residues
(thinking that dihedral potential affects 4 consecutive residues).

It also save a list "chains_fixed.csv" where it is reported
if a chain has been modeled and the segments modeled.

Note about res numeration:
It has been re-numbered starting from 1.
Consistent and useful for CATH domain definition.

POSSIBLE ISSUES:
A possible bug is that, for some pdbs, MODRES residues are considered
as missing residues. A check with ATOM records find the discrepancy.
This could be a "problem" with modeller that reads only ATOM records.
"""
import csv
import os
import pathlib
import shutil
import sys
import time
from dataclasses import dataclass

import modeller
import numpy as np
from Bio import SeqIO
from Bio.PDB import parse_pdb_header
from modeller import Selection
from modeller.automodel import AutoModel

# read input from command line to analyse only a batch
# of the whole chain_list dataset
if len(sys.argv) > 1:
    batch_start = int(sys.argv[1])
    if sys.argv[2] == 'end':
        batch_end = 'end'
    else:
        batch_end = int(sys.argv[2])
    batch_num_id = int(sys.argv[3])


@dataclass
class MisRes:
    """Contains information about a residue in a REMARK 465 record.

    The insertion code, e.g. 'A'. The insertion code is sometimes
    used to preserve a certain desirable residue numbering scheme.
    A Ser 80 insertion mutant (inserted e.g. between a Thr 80 and an Asn
    81 residue) could e.g. have sequence identifiers and insertion codes
    as follows: Thr 80 A, Ser 80 B, Asn 81.
    In this way the residue numbering scheme stays in tune with that
    of the wild type structure.

    Insertion can be None or a string
    """

    ssseq: int
    insertion: str

    def has_insertion(self):
        """Return True if the residue has an insertion code."""
        return self.insertion is not None


start_time = time.time()

modeller.log.none()  # no prints

parent = pathlib.Path(__file__).parent.resolve()
membrane_dir = pathlib.Path("/mnt/MEMBRANE/")
chain2fix_dir = membrane_dir / "CHAINS"
alignments_dir = membrane_dir / "ALIGNMENTS"
chain_fixed_dir = membrane_dir / "CHAINS_FIX"
log_dir = chain_fixed_dir / "LOG"

# save chain that has been fix and those that have been skipped
chain_fixed_path = log_dir / f"chains_fixed_{batch_num_id}.csv"
chain_fixed_file = open(chain_fixed_path, "w", encoding="utf-8")
chain_fixed_writer = csv.writer(chain_fixed_file)
chain_fixed_writer.writerows(
    [
        ["# chain_id", "fixed", "residues_missing", "modeled_residues"],
        ["# residues_missing -> those extracted from the REMARK 465 records"],
        ["# and with id re-numberated according to alignment"],
        ["# modeled_residues -> those adjusted by modeller; not fixed"]
    ]
)

chain_list_file = membrane_dir / "chains_list.csv"
with open(chain_list_file, "r", encoding="utf-8") as infile:
    reader = csv.reader(infile)
    chain_list = list(reader)

if batch_end == 'end':
    batch_end = len(chain_list)
for row in chain_list[batch_start:batch_end]:
    chain_name = row[2]
    chain_id = f"{row[0][:4]}_{row[2]}"
    pdb_path = str(chain2fix_dir / f"{chain_id}.pdb")

    # Check through REMARK 465 records if there are missing residues
    remark_465_found = False
    pdb_header = parse_pdb_header(pdb_path)
    if "has_missing_residues" in pdb_header:
        missing_residues = [
            res for res in pdb_header["missing_residues"] if res["chain"] == chain_name
        ]
        if missing_residues:
            # In this cas, there are residue missing in the chain of
            # interest. Hence, we proceed with the modelling
            pass
        else:
            # In this case, there are no missing residues in the chain
            # of interest. Hence, we skip the modelling
            chain_fixed_writer.writerow([chain_id, "no"])
            # Copy the chain to the CHAINS_FIX folder
            shutil.copyfile(
                chain2fix_dir / f"{chain_id}.pdb",
                chain_fixed_dir / f"{chain_id}.pdb")
            continue

    # Modeller env and model of the structure to complete
    env = modeller.Environ()
    modeller.log.none()
    env.io.atom_files_directory = [str(chain2fix_dir)]
    model = modeller.Model(env, file=chain_id)

    # save sequence to complete (obtain from ATOM records)
    alignment = modeller.Alignment(env)
    alignment.append_model(model, align_codes=chain_id)
    alignment.write(file=str(alignments_dir / f"{chain_id}.atom.seq"))

    # save sequence to complete (obtain from SEQRES records)
    seqres_found = False
    for record in SeqIO.parse(pdb_path, "pdb-seqres"):
        if record.id[-1:] == chain_name:
            seqres_found = True
            record.id = f"{chain_id}_complete"
            out_seqres = str(alignments_dir / f"{chain_id}.seqres.fasta")
            SeqIO.write(record, out_seqres, "fasta")
    if not seqres_found:
        raise ValueError(f"SEQRES record not found for chain {chain_id}")

    # Align complete sequence with ATOM sequence
    alignment.append(file=out_seqres, alignment_format="fasta")
    alignment.salign()  # perform the actual alignment
    out_ali = str(alignments_dir / f"{chain_id}.ali")
    alignment.write(
        file=out_ali, alignment_format="PIR"
    )
    alignment.check()  # additional step to control alignment

    # Now, we exploit the alignment to renumber missing residues
    # read from the REMARK 465 records
    seq_w_gaps = SeqIO.to_dict(SeqIO.parse(out_ali, "pir"))[chain_id].seq
    if seq_w_gaps.startswith("-"):
        # the first residue is missing and correspond to the first
        # residue in the REMARK 465 record
        pass
    seq_w_gaps_list = np.array([letter for letter in seq_w_gaps])
    n_sequence_w_gaps = len(seq_w_gaps_list)
    # +1 to match res numeration
    missing_residues_idx = np.where(seq_w_gaps_list == "-")[0] + 1
    # create a list wit starting and finishing missing segments
    segments = []
    start_seg = missing_residues_idx[0]
    increment = 1
    num_missing_residues_2 = len(missing_residues_idx[1:]) - 1
    # WARN: we add 3 residues to be moved in order to take into account
    # dihedral potential and void prohibit dihedrals in the modeled residues
    # Only one missing residue
    if num_missing_residues_2 == -1:
        if start_seg == 1:
            # N terminal
            segments.append((start_seg, start_seg + 3))
        else:
            # C terminal
            segments.append((start_seg - 3, start_seg))
    else:
        for idx, mis_res in enumerate(missing_residues_idx[1:]):
            if mis_res == start_seg + increment and idx != num_missing_residues_2:
                increment += 1
            else:
                if start_seg >= 4:
                    append2N = - 3
                else:
                    append2N = - (start_seg - 1)
                if start_seg + increment <= n_sequence_w_gaps - 3:
                    if idx == num_missing_residues_2:
                        append2C = 3
                    else: 
                        append2C = 2  # to compensate the +1 in the increment
                else:
                    append2C = n_sequence_w_gaps - (start_seg + increment)
                segments.append(
                    (start_seg + append2N, start_seg + increment + append2C)
                )
                if idx == num_missing_residues_2:
                    break
                start_seg = mis_res
                increment = 1
    if not segments:
        raise RuntimeError(
            (
                f"No segments found for chain {chain_id}\n"
                f"missing_residues_idx: {missing_residues_idx}"
            )
        )
    # save modelled residues
    chain_fixed_writer.writerow(
        [chain_id, "yes", missing_residues_idx, segments]
    )
    
    # Define class for modeller
    class ModelSegment(AutoModel):
        """AutoModel class but only for missing residues.
        
        References:
        - https://salilab.org/modeller/wiki/Missing_residues
        - https://salilab.org/modeller/10.3/manual/node30.html
        """

        def select_atoms(self):
            """Select residues to be refined.
            
            Residues not selected will be kept fixed.
            """
            return Selection(
                *[self.residue_range(f"{start}:{chain_name}", f"{end}:{chain_name}") for start, end in segments]
            )
        
        def special_patches(self, aln):
            """Rename chains and renumber residues.
            
            Explicitly force to renumber residues from 1.
            NOTE: although aln is not used, it is required by the
            internal methods implemented by modeller
            """
            self.rename_segments(
                segment_ids=[chain_name],
                renumber_residues=[1]
            )

    loopmodel = ModelSegment(
        env,
        alnfile=str(alignments_dir / f"{chain_id}.ali"),
        knowns=chain_id,  # sequence to be completed
        sequence=f"{chain_id}_complete",  # target (complete) sequence
    )
    # the following force to build only one model
    loopmodel.starting_model = 1
    loopmodel.ending_model = 1

    print(f"===== Modelling chain {chain_id} =====")
    print(f"Number of total residues: {n_sequence_w_gaps}")
    print(f"Missing residues: {missing_residues_idx}")
    print(f"Segments: {segments}")
    loopmodel.make()

    # check if run was successful
    if loopmodel.outputs[0]["failure"] is None:
        print("========== Success! ==========\n")
    # Save output file name and rename it
    final_output_file = os.getcwd() + "/" + loopmodel.outputs[0]["name"]
    shutil.copyfile(final_output_file, chain_fixed_dir / f"{chain_id}.pdb")
    # Remove output files
    files2remove = [
        os.getcwd() + "/" + f for f in os.listdir(os.getcwd()) if f.startswith(f"{chain_id}_complete")
    ]
    for file in files2remove:
        os.remove(file)

chain_fixed_file.close()

print(f"Elapsed time: {time.time() - start_time:.2f} s")
